"""XYZ D65 class."""
from ..spaces import RE_DEFAULT_MATCH, GamutUnbound
from . import _cat
from .xyz import XYZ
import re


class XYZD65(XYZ):
    """XYZ D65 class."""

    SPACE = "xyz-d65"
    CHANNEL_NAMES = ("x", "y", "z", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = _cat.WHITES["D65"]

    _range = (
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0])
    )

    @classmethod
    def _to_xyz(cls, xyzd65):
        """To XYZ."""

        return _cat.chromatic_adaption(cls.white(), XYZ.white(), xyzd65)

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return _cat.chromatic_adaption(XYZ.white(), cls.white(), xyz)



"""Lch D65 class."""
from ..spaces import RE_DEFAULT_MATCH
from . import _cat
from .lab_d65 import LabD65
from .lch import LchBase, lch_to_lab, lab_to_lch
import re


class LchD65(LchBase):
    """Lch D65 class."""

    SPACE = "lch-d65"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = _cat.WHITES["D65"]

    @classmethod
    def _to_lab_d65(cls, lchd65):
        """To Lab."""

        return lch_to_lab(lchd65)

    @classmethod
    def _from_lab_d65(cls, labd65):
        """To Lab."""

        return lab_to_lch(labd65)

    @classmethod
    def _to_xyz(cls, lch):
        """To XYZ."""

        return LabD65._to_xyz(cls._to_lab_d65(lch))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return cls._from_lab_d65(LabD65._from_xyz(xyz))



"""Convert utilities."""
from .. import util


class Convert:
    """Conversion methods."""

    def convert(self, space, *, fit=False, in_place=False):
        """Convert to color space."""

        space = space.lower()

        if fit:
            method = None if not isinstance(fit, str) else fit
            if not self.in_gamut(space, tolerance=0.0):
                converted = self.convert(space, in_place=in_place)
                converted.fit(space, method=method, in_place=True)
                return converted

        convert_to = '_to_{}'.format(space)
        convert_from = '_from_{}'.format(self.space())

        obj = self.CS_MAP.get(space)
        if obj is None:
            raise ValueError("'{}' is not a valid color space".format(space))

        # See if there is a direct conversion route
        func = None
        coords = self.coords()
        if hasattr(self._space, convert_to):
            func = getattr(self._space, convert_to)
            coords = func(coords)
        elif hasattr(obj, convert_from):
            func = getattr(obj, convert_from)
            coords = func(coords)

        # See if there is an XYZ route
        if func is None and self.space() != space:
            func = getattr(self._space, '_to_xyz')
            coords = func(coords)

            if space != 'xyz':
                func = getattr(obj, '_from_xyz')
                coords = func(coords)

        converted = self.new(space, coords, self.alpha)

        return self.mutate(converted) if in_place else converted

    def mutate(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Mutate the current color to a new color."""

        self._attach(self._parse(color, data, alpha, filters=filters, **kwargs))
        return self

    def update(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Update the existing color space with the provided color."""

        clone = self.clone()
        obj = self._parse(color, data, alpha, filters=filters, **kwargs)
        clone._attach(obj)

        if clone.space() != self.space():
            clone.convert(self.space(), in_place=True)

        self._attach(clone._space)
        return self



"""Lab class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Percent
from . import _cat
from .xyz import XYZ
from .. import util
import re

EPSILON3 = 216 / 24389  # `6^3 / 29^3`
EPSILON = 24 / 116
RATIO1 = 16 / 116
RATIO2 = 108 / 841
RATIO3 = 841 / 108


def lab_to_xyz(lab):
    """
    Convert Lab to D50-adapted XYZ.

    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    l, a, b = lab

    # compute `f`, starting with the luminance-related term
    fy = (l + 16) / 116
    fx = a / 500 + fy
    fz = fy - b / 200

    # compute `xyz`
    xyz = [
        fx ** 3 if fx > EPSILON else (fx - RATIO1) * RATIO2,
        fy ** 3 if fy > EPSILON or l > 8 else (fy - RATIO1) * RATIO2,
        fz ** 3 if fz > EPSILON else (fz - RATIO1) * RATIO2
    ]

    # Compute XYZ by scaling `xyz` by reference `white`
    return util.multiply(xyz, Lab.white())


def xyz_to_lab(xyz):
    """Assuming XYZ is relative to D50, convert to CIE Lab from CIE standard."""

    # compute `xyz`, which is XYZ scaled relative to reference white
    xyz = util.divide(xyz, Lab.white())
    # Compute `fx`, `fy`, and `fz`
    fx, fy, fz = [util.cbrt(i) if i > EPSILON3 else (RATIO3 * i) + RATIO1 for i in xyz]

    return (
        (116.0 * fy) - 16.0,
        500.0 * (fx - fy),
        200.0 * (fy - fz)
    )


class LabBase(Space):
    """Lab class."""

    CHANNEL_NAMES = ("lightness", "a", "b", "alpha")

    _range = (
        GamutUnbound([Percent(0), Percent(100.0)]),  # Technically we could/should clamp the zero side.
        GamutUnbound([-160, 160]),  # No limit, but we could impose one +/-160?
        GamutUnbound([-160, 160])  # No limit, but we could impose one +/-160?
    )

    @property
    def lightness(self):
        """L channel."""

        return self._coords[0]

    @lightness.setter
    def lightness(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def a(self):
        """A channel."""

        return self._coords[1]

    @a.setter
    def a(self, value):
        """A axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self):
        """B channel."""

        return self._coords[2]

    @b.setter
    def b(self, value):
        """B axis."""

        self._coords[2] = self._handle_input(value)


class Lab(LabBase):
    """Lab class."""

    SPACE = "lab"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = _cat.WHITES["D50"]

    @classmethod
    def _to_xyz(cls, lab):
        """To XYZ."""

        return _cat.chromatic_adaption(cls.white(), XYZ.white(), lab_to_xyz(lab))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return xyz_to_lab(_cat.chromatic_adaption(XYZ.white(), cls.white(), xyz))



"""Lab D65 class."""
from ..spaces import RE_DEFAULT_MATCH
from . import _cat
from .xyz import XYZ
from .lab import LabBase, lab_to_xyz, xyz_to_lab
import re


class LabD65(LabBase):
    """Lab D65 class."""

    SPACE = "lab-d65"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    WHITE = _cat.WHITES["D65"]

    @classmethod
    def _to_xyz(cls, lab):
        """To XYZ."""

        return _cat.chromatic_adaption(cls.white(), XYZ.white(), lab_to_xyz(lab))

    @classmethod
    def _from_xyz(cls, xyz):
        """From XYZ."""

        return xyz_to_lab(_cat.chromatic_adaption(XYZ.white(), cls.white(), xyz))



"""HWB class."""
from ..spaces import Space, RE_DEFAULT_MATCH, Angle, Percent, GamutBound, Cylindrical
from . import _cat
from .srgb import SRGB
from .hsv import HSV
from .. import util
import re


def hwb_to_hsv(hwb):
    """HWB to HSV."""

    h, w, b = hwb
    w /= 100.0
    b /= 100.0

    wb = w + b
    if (wb >= 1):
        gray = w / wb
        return [util.NaN, 0.0, gray * 100.0]

    v = 1 - b
    s = 0 if v == 0 else 1 - w / v
    return [h, s * 100, v * 100]


def hsv_to_hwb(hsv):
    """HSV to HWB."""

    h, s, v = hsv
    s /= 100
    v /= 100
    w = v * (1 - s)
    b = 1 - v
    if w + b >= 1:
        h = util.NaN
    return [h, w * 100, b * 100]


class HWB(Cylindrical, Space):
    """HWB class."""

    SPACE = "hwb"
    CHANNEL_NAMES = ("hue", "whiteness", "blackness", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE))
    GAMUT_CHECK = "hsl"
    WHITE = _cat.WHITES["D65"]

    _range = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    @property
    def hue(self):
        """Hue channel."""

        return self._coords[0]

    @hue.setter
    def hue(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def whiteness(self):
        """Whiteness channel."""

        return self._coords[1]

    @whiteness.setter
    def whiteness(self, value):
        """Set whiteness channel."""

        self._coords[1] = self._handle_input(value)

    @property
    def blackness(self):
        """Blackness channel."""

        return self._coords[2]

    @blackness.setter
    def blackness(self, value):
        """Set blackness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] + coords[2] >= 100:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def _to_xyz(cls, hwb):
        """SRGB to XYZ."""

        return SRGB._to_xyz(cls._to_srgb(hwb))

    @classmethod
    def _from_xyz(cls, xyz):
        """XYZ to SRGB."""

        return cls._from_srgb(SRGB._from_xyz(xyz))

    @classmethod
    def _to_srgb(cls, hwb):
        """To sRGB."""

        return HSV._to_srgb(cls._to_hsv(hwb))

    @classmethod
    def _from_srgb(cls, srgb):
        """From sRGB."""

        return cls._from_hsv(HSV._from_srgb(srgb))

    @classmethod
    def _to_hsl(cls, hwb):
        """To HSL."""

        return HSV._to_hsl(hwb_to_hsv(hwb))

    @classmethod
    def _from_hsl(cls, hsl):
        """From HSL."""

        return hsv_to_hwb(HSV._from_hsl(hsl))

    @classmethod
    def _to_hsv(cls, hwb):
        """To HSV."""

        return hwb_to_hsv(hwb)

    @classmethod
    def _from_hsv(cls, hsv):
        """From HSV."""

        return hsv_to_hwb(hsv)



"""Test distance methods."""
from coloraide import Color
from . import util
import pytest


class TestDistance(util.ColorAsserts):
    """Test distance."""

    def assertEqual(self, a, b):
        """Assert equal."""

        assert a == b

    def assertNotEqual(self, a, b):
        """Assert not equal."""

        assert a != b

    @pytest.mark.parametrize(
        'color1,color2,value',
        [
            ('red', 'red', 0),
            ('red', 'orange', 137.1861),
            ('red', 'yellow', 207.1759),
            ('red', 'green', 238.3253),
            ('red', 'blue', 349.7736),
            ('red', 'indigo', 260.9437),
            ('red', 'violet', 188.974),
            ('red', 'white', 232.6692),
            ('red', 'black', 370.0502),
            ('red', 'gray', 205.0097),
            ('red', 'red', 0),
            ('orange', 'red', 137.1861),
            ('yellow', 'red', 207.1759),
            ('green', 'red', 238.3253),
            ('blue', 'red', 349.7736),
            ('indigo', 'red', 260.9437),
            ('violet', 'red', 188.974),
            ('white', 'red', 232.6692),
            ('black', 'red', 370.0502),
            ('gray', 'red', 205.0097)
        ]
    )
    def test_delta_e_itp(self, color1, color2, value):
        """Test delta e ITP."""

        print('color1: ', color1)
        print('color2: ', color2)
        self.assertCompare(
            Color(color1).delta_e(color2, method="itp"),
            value,
            rounding=4
        )

    @pytest.mark.parametrize(
        'color1,color2,value',
        [
            ('red', 'red', 0),
            ('red', 'orange', 45.7287),
            ('red', 'yellow', 69.0586),
            ('red', 'green', 79.4418),
            ('red', 'blue', 116.5912),
            ('red', 'indigo', 86.9812),
            ('red', 'violet', 62.9913),
            ('red', 'white', 77.5564),
            ('red', 'black', 123.3501),
            ('red', 'gray', 68.3366),
            ('red', 'red', 0),
            ('orange', 'red', 45.7287),
            ('yellow', 'red', 69.0586),
            ('green', 'red', 79.4418),
            ('blue', 'red', 116.5912),
            ('indigo', 'red', 86.9812),
            ('violet', 'red', 62.9913),
            ('white', 'red', 77.5564),
            ('black', 'red', 123.3501),
            ('gray', 'red', 68.3366)
        ]
    )
    def test_delta_e_itp_2000(self, color1, color2, value):
        """
        Test delta e ITP with scalar of 240.

        240 is supposed to give an equivalence to the average delta E 2000 results.
        """

        print('color1: ', color1)
        print('color2: ', color2)
        self.assertCompare(
            Color(color1).delta_e(color2, method="itp", scalar=240),
            value,
            rounding=4
        )



"""Fit by compressing chroma in Lch."""


def fit(base, color):
    """
    Gamut mapping via chroma Lch.

    Algorithm comes from https://colorjs.io/docs/gamut-mapping.html.

    The idea is to hold hue and lightness constant and decrease lightness until
    color comes under gamut.

    We'll use a binary search and at after each stage, we will clip the color
    and compare the distance of the two colors (clipped and current color via binary search).
    If the distance is less than two, we can return the clipped color.

    ---
    Original Authors: Lea Verou, Chris Lilley
    License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
    """

    # Compare clipped against original to
    # judge how far we are off with the worst case fitting
    space = color.space()
    clipped = color.clone()
    clipped.fit(space=space, method="clip", in_place=True)
    base_error = base.delta_e(clipped, method="2000")

    if base_error > 2.3:
        threshold = .001
        # Compare mapped against desired space
        mapcolor = color.convert("lch")
        error = color.delta_e(mapcolor, method="2000")
        low = 0.0
        high = mapcolor.chroma

        # Adjust chroma (using binary search).
        # This helps preserve the color more (in most cases).
        # After each adjustment, see if clipping gets us close enough.
        while (high - low) > threshold and error < base_error:
            clipped = mapcolor.clone()
            clipped.fit(space, method="clip", in_place=True)
            delta = mapcolor.delta_e(clipped, method="2000")
            error = color.delta_e(mapcolor, method="2000")
            if delta - 2 < threshold:
                low = mapcolor.chroma
            else:
                if abs(delta - 2) < threshold:  # pragma: no cover
                    # Can this occur?
                    break
                high = mapcolor.chroma
            mapcolor.chroma = (high + low) / 2
        # Trim off noise allowed by our tolerance
        color.update(mapcolor)
        color.fit(space, method="clip", in_place=True)
    else:
        # We are close enough that we should just clip.
        color.update(clipped)
    return color.coords()
